##########################################################################################
##########################################################################################
# PowerShell Goodness 2017
# Useful tips and tricks figured out and found on the web...
# NOTE: This file contains examples only; it is not intended to be run as a single script.
#
# Jeremy Sublett
# jsublett@composablesystems.com
# Composable Systems, LLC
# http://composablesystems.com
#
##########################################################################################
##########################################################################################

#region Basics

##########################################################################################
# Basics
##########################################################################################

#region Basic basics

dir
cls
cd

#output
write "test"
Write-Output "Hello... output"
Write-Host "Hello... host"
Write-Host "Hello... host with fancy color." -BackgroundColor Yellow

Write-Warning "This is an error message."

Write-Debug "blah"

Get-Command
Get-Command *get-*
command

Show-Command Get-Command


#using tabbing
#get-co [tab]
#*com [tab]

#HELP
Help Get-Command
Help Get-Command -Examples

Update-Help

#retrieving multiple items
dir
Get-ChildItem

#GET CHILD ITEMS -- from other locations
Get-ChildItem registry::hklm\software
dir registry::hklm\software
Get-ChildItem cert:\. -recurse -codesigningcert

Get-ChildItem Env:
dir $env:windir
dir $env:TEMP

#endregion

#region piping

Get-Command | Select-Object Name, CommandType

Get-Command | Group-Object CommandType

command | sort CommandType | select Name, CommandType

Get-Command | Format-List
Get-Command | Format-Table
Get-Command | Format-Wide -AutoSize 

#output to file
Get-Command | Out-File commandoutput.txt

Get-Command | Select-Object name | Out-File commandoutput.txt

#output to grid
Get-Command | Out-GridView

#CSV output
Get-Command | export-csv Commands.csv

#XML output
(Get-Command | ConvertTo-XML).Save("c:\Commands.xml")

#HTML output
Get-Command | ConvertTo-Html -title "Commands" -body "<H2>Commands.</H2>" | Set-Content c:\temp\commands.htm

#grid output...
Get-Command | Out-GridView

#endregion

#region variables

$x = 10
$x = $y + $z
$x = $y

#swap values
$x,$y = $y,$x

#bulk initialization
$x,$y,$z = 10,20,30

$a = Get-Command
$a
$a | select CommandType, Name
$b = dir

$a = Read-Host "Please enter a value"

# '&' operator -- allows string to be treated as a command
$x = "Get-Command"
$x
&$x

$myarray = "red","green","blue"
"red" -in $myarray

$numbers = 1..40
50 -notin $numbers

10 / 20
"10" / 20

[int]$x = 10

$x = 4
$x = "test"  #Error

$loc = Get-Location
$file = Get-ChildItem "$loc\Stock Symbols.csv"

#endregion

#region filtering with Where-Object / ?

dir | Where-Object { $_.Name -like "*.txt" }

dir | ? { $_.Name -like "*.txt" }
 
dir | where Name -like "*.txt"   # New in PS 3

#operators... -and -or -like

dir | where { $_.Name -like "*.txt" -and $_.Length -lt 7000}
 
Get-Command | Where { $_.name -like "*trans*" }

Get-WmiObject -list | where {$_.name -match "Network"}

#endregion

#endregion Basics

#region Advanced Basics

##########################################################################################
# Advanced Basics
##########################################################################################


#input
$a = Read-Host "Please enter a value"


#parameters to a script file

param([Int32]$myVar=30) #Must be the first statement in your script
#...
#how to call from cmd line: powershell.exe -file myScript.ps1 -myVar 15
#how to call from powershell: myScript.ps1 -myVar 15


#using regular expressions

dir $home -Recurse | Where-Object {$_.Name -match '\d.*?\.'}

#region functions

function Hello-World
{
    $day = (Get-Date).DayOfWeek
    Write-Output "Hello. Today is $day."
}

#parameters
function Add-Numbers([int]$x,[int]$y)
{
	return ($x + $y)
}
Add-Numbers 4 5


#using the $input variable from the pipe
function Echo-Pipe
{
 	return $input
}
Add-Numbers 20 30 | Echo-Pipe


function FindXmlFiles
{
 	$input | Where-Object {$_.Extension -eq ".xml"}
}
dir -Path C:\Windows | FindXmlFiles


function Out-ReportExcel {
 param(
  $Path = "$env:temp\report$(Get-Date -format yyyyMMddHHmmss).csv",
  [switch]$Open
 )

 $Input | 
  Export-Csv $Path -NoTypeInformation -UseCulture -Encoding UTF8

  if($Open) { Invoke-Item $Path }
}
Get-Process | Out-ReportExcel -Open

#endregion

#region modules

#in a .psm1 file...
#begin file: sample.psm1
$neighborhood = "some place"

function New-House ($address, $roomcount, $bathroomcount) {
  # do something
}

function Get-House ($address) {
  #do something
}

function Set-House ($address, $roomcount, $bathroomcount) {
  #do something
}

function Remove-House ($address) {
  #do something
}

Export-ModuleMember -Function "Get-*"
Export-ModuleMember -Function "Set-*"
Export-ModuleMember -Variable "Neighborhood"
#end file

#then import the module in your ps script to make the functions/variable available to you
Import-Module sample.psm1

#endregion modules

#region looping

#for
For ($a = 1; $a -le 10; $a++) { Write-Host $a }

For ($i=1; $i -le 100; $i++)
{
  Start-Sleep -Milliseconds 1
  Write-Progress -Activity "Doing some work" -Status "Progress:" -PercentComplete (($i/100)*100)
}

#foreach / %
$ary = (1,2,3,4,5,6,7,8,9,10,11,12)
Foreach ($item in $ary) {$item * 100}

#foreach, piped
$ary | foreach ($_ * 100)

# the % operator
####% ($item in $ary) {$item * 500}


#while
$i =8 
While ($i -le 96) {$i; $i +=8}

$procs=Get-Process | select-object ProcessName
$i=1
foreach ($process in $procs)
{
	Write-Host $i "Process Name is" $process.Processname;
	$i++;
}



#looping through the stock symbols

$csv = Import-Csv "Stock symbols.csv"
$csv | foreach { $_.Symbol }
#lookup each quote
$csv | foreach { Get-Quote $_.Symbol }

#endregion

#region working with JSON data

#get data from json string
$employeeJson = '{"employees":[
    {"firstName":"John", "lastName":"Doe"},
    {"firstName":"Anna", "lastName":"Smith"},
    {"firstName":"Peter", "lastName":"Jones"}
]}'

$employeeObj = ConvertFrom-Json $employeeJson

$employeeObj.employees | select firstname

#generate JSON from an object

Get-ChildItem | ConvertTo-Json

#endregion

#region classes and enumerations

Enum Animal {
    Horse
    Cat
    Dog
    Mouse
}

Enum Speed {
    Fast = 100
    Medium = 50
    Slow = 20
    Stopped = 0
}

Class Racer
{
    [Animal]$animal = [Animal]::Horse
    [string]$name
    [Speed]$speed

    Racer()
    {
    }
}

$x = [Racer]::new()

#endregion

#endregion Advanced Basics

#region Web and REST APIs

##########################################################################################
# Web and REST APIs
##########################################################################################

#region Simple HTML

Invoke-WebRequest -Method Get -Uri http://microsoft.com

$postParams = @{username='me';moredata='qwerty'}
Invoke-WebRequest -Uri http://example.com/foobar -Method POST -Body $postParams

#endregion

#region REST APIs

#sample api site: http://jsonplaceholder.typicode.com/

#using Invoke-RestMethod
$apiResults = Invoke-RestMethod -Method Get -Uri http://jsonplaceholder.typicode.com/posts/1/comments

#using Invoke-WebRequest (have to convert to JSON)
$x = Invoke-WebRequest -Method Get -Uri http://jsonplaceholder.typicode.com/posts/1/comments
$apiResults = $x.Content | ConvertFrom-Json

# posting data
Invoke-RestMethod -Method Post -Uri $url -Body (ConvertTo-Json $body) -Header @{"X-ApiKey"=$apiKey}

#endregion

#region SOAP APIs "Web Services"

$svc = New-WebServiceProxy -Uri http://www.webservicex.net/CurrencyConvertor.asmx?WSDL

#get the current conversion rate
$svc.ConversionRate("USD","EUR")

#interrogate with Get-Member
$svc.ConversionRate | Get-Member

function ConvertTo-Euro
{
	param ([int]$Dollars)
	
	$svc = New-WebServiceProxy -Uri http://www.webservicex.net/CurrencyConvertor.asmx?WSDL

	$rate = $svc.ConversionRate('USD','EUR')
	
	$Euros = $Dollars * $rate
	
	Write-Host ([double]$Euros)
}

ConvertTo-Euro 45.00


function Get-Quote
{
	param ([string]$symbol)

	$svc = New-WebServiceProxy -Uri http://www.webservicex.net/stockquote.asmx

	[xml]$results = $svc.GetQuote($symbol)
	
	Write-Host ([double]$results.StockQuotes.Stock.Last)
}

Get-Quote "MSFT"
Get-Quote "YHOO"
Get-Quote "TXRH"

#endregion

#endregion

#region Office 365

##########################################################################################
# Office 365
##########################################################################################

#region Office 365: Management Azure AD
#http://technet.microsoft.com/en-us/library/jj151815.aspx#BKMK_connect

$cred = Get-Credential
Connect-MsolService -Credential $cred

Get-MsolCompanyInformation

$LoginId = Read-Host -Prompt "Enter login id."
Get-MsolUser -UserPrincipalName $LoginId

Get-MsolRole | Select Name, Description

#endregion

#region Office 365: Exchange

$session = New-PSSession -ConfigurationName MicrosoftExchange -ConnectionUri https://ps.outlook.com/powershell/ -Credential $cred -Authentication Basic -AllowRedirection

Import-PSSession $Session

Remove-PSSession $Session

#endregion

#region Office 365: SharePoint Online Management

$AdminUrl = Read-Host -Prompt "Enter the SPO admin URL. ex: https://acme-admin.sharepoint.com"
$LoginId = Read-Host -Prompt "Enter login id."
Connect-SPOService -url $AdminUrl -Credential $LoginId

Get-SPOSite

$SiteCollectionUrl = Read-Host -Prompt "Enter site collection URL."
Get-SPOSiteGroup -Site $SiteCollectionUrl | select title

Get-SPOUser -Site $SiteCollectionUrl -LoginName $LoginId

#endregion

#endregion Office 365

#region SQL Server

##########################################################################################
# SQL Server
##########################################################################################

#https://northwinddatabase.codeplex.com/

Invoke-Sqlcmd -Query "SELECT @@VERSION;" 

Invoke-Sqlcmd -ServerInstance . -Database Northwind -Query "SELECT * FROM Employees"

Invoke-Sqlcmd -ServerInstance . -Database Northwind -Query "SELECT * FROM Employees" | Select FirstName, EmployeeID

Invoke-Sqlcmd -ServerInstance . -Database Northwind -Query "SELECT * FROM Employees" | Out-GridView


function Get-EmployeeCount() {
  $x = Invoke-Sqlcmd -ServerInstance . -Database Northwind -Query "SELECT COUNT(*) FROM Employees"
  return ($x)[0]
}

function Get-Employee([int]$emplId)
{
  return Invoke-Sqlcmd -ServerInstance . -Database Northwind -Query "SELECT * FROM Employees WHERE EmployeeID = $emplId"
}

#directing output...
$x = Invoke-Sqlcmd -ServerInstance . -Database Northwind -Query "SELECT * FROM Employees" | Select FirstName, EmployeeID

#file output
$x | Out-File c:\Employees.txt

#endregion

#region SharePoint

##########################################################################################
# SharePoint
##########################################################################################

#note: this was used with SharePoint on-prem
$snapin = Get-PSSnapin | Where-Object { $_.Name -eq "Microsoft.SharePoint.Powershell" }
if ($snapin -eq $null)
{
    Write-Host "Loading SharePoint Powershell Snapin..."
    Add-PSSnapin "Microsoft.SharePoint.Powershell"
}

#adding a new site collection
$siteColl = New-SPSite -URL "http://csdev/sites/demo1" -OwnerAlias JSE6410\Jeremy -Template "STS#1" -Name "Demo 1"

#creating a document library
$csv = Import-Csv c:\PowerShellDemo\symbols.csv
$csv | foreach { Get-Quote $_.Symbol }

#creating a bunch of doc libraries - based on data in SQL
$vendors = Invoke-Sqlcmd -Database AdventureWorks -Query "Select top 10 * from Purchasing.Vendor"
$vendors | foreach { $siteColl.RootWeb.Lists.Add($_.AccountNumber,$_.AccountNumber,$listTemplate) }

#cleanup
$siteColl.Dispose()

#removing
Remove-SPSite -Identity "http://csdev/sites/demo1" -Confirm:$false

#region IISReset Across SharePoint Farm

Write-Host "Loading SharePoint Commandlets"
Add-PSSnapin Microsoft.SharePoint.PowerShell -erroraction SilentlyContinue
Write-Host -ForegroundColor Green " Commandlets Loaded ... Loading Variables"
Write-Host
[array]$servers= Get-SPServer | ? {$_.Role -eq "Application"}
$farm = Get-SPFarm
foreach ($server in $servers)
{
     Write-Host -ForegroundColor Yellow "Attempting to reset IIS for $server"
        iisreset $server /noforce "\\"$_.Address
        iisreset $server /status "\\"$_.Address
        Write-Host
        Write-Host -ForegroundColor Green "IIS has been reset for $server"
        Write-Host
}
Write-Host -ForegroundColor Green "IIS has been reset across the SharePoint Farm"
Start-Sleep -Seconds 5
Write-host

#endregion

#endregion

#region Extra

##########################################################################################
# Extra
##########################################################################################

#region progress bar

#simple
for ($i = 1; $i -le 10; $i++ )
{
  sleep(1)
  $pct = $i * 10
  Write-Progress -Activity "Search in Progress" -Status "$pct% Complete:" -PercentComplete $pct;
}

#display during operation
$Events = Get-EventLog -LogName system
$Events | ForEach-Object -Begin { Clear-Host;$I=0;$out="" } -Process {
    if ($_.message -like "*bios*")
        { $out=$out + $_.Message };
        $I = $I+1;
        Write-Progress -Activity "Searching Events" -Status "Progress:" -PercentComplete ($I/$Events.count*100)
} -End {$out}

#endregion

#region Custom HTML Reporting

#creates a report about this system in a HTML file
function Get-SystemReport($ComputerName = $env:ComputerName) {
  $htmlStart = "
  <HTML><HEAD><TITLE>Server Report</TITLE>
  <style>  
  body { background-color:#EEEEEE; }  
  body,table,td,th { font-family:Tahoma; color:Black; Font-Size:10pt }  
  th { font-weight:bold; background-color:#AAAAAA; }  
  td { background-color:white; }  
  </style></HEAD><BODY>
  <h2>Report listing for System $Computername</h2>
  <p>Generated $(get-date -Format 'yyyy-MM-dd hh:mm') </p>
  "
  
  $htmlEnd = '</body></html>'
  
    $htmlStart 
    Get-WmiObject -Class CIM_PhysicalElement | 
      Group-Object -Property __Class | 
      ForEach-Object { 
        $_.Group |
        Select-Object -Property * | ConvertTo-HTML -Fragment -PreContent ('<h3>{0}</h3>' -f $_.Name ) 
    }
    $htmlEnd 
}

$path = "$env:temp\report.htm"
Get-SystemReport | Out-File -Filepath $path 
Invoke-Item $path

#endregion

#region Event Log

Get-EventLog -LogName 'Application' -Newest 100 -EntryType Error

Get-EventLog -LogName 'Application' -Newest 100 -EntryType Error | Select Message

#on a different computer
Get-EventLog -ComputerName Server123 -LogName 'Application' -Newest 100 -EntryType Error

#TIP - limit results in base cmdlet before filtering in PowerShell
Get-EventLog -LogName 'Application' -Newest 100 -EntryType Error | where { $_.Message -like '*25*' }
#NOTE: above returned 100 to PowerShell, then it's filtered down

#endregion

#region workflow

workflow Test-Workflow
{
    Parallel
    {
         Get-Process
         Get-Service
    }
}


function Do-Something1
{
    sleep 4
    "done 1"
}

function Do-Something2
{
    sleep 3
    "done 2"
}

function Do-Something3
{
    sleep 2
    "done 3"
}

function Do-Something4
{
    sleep 1
    "done 4"
}

workflow Test-Workflow
{
    Parallel
    {
         Do-Something1
         Do-Something2
         Do-Something3
         Do-Something4
    }
}


#endregion

#region Message box

$msg = New-Object -ComObject WScript.Shell
$msg.Popup("Hello", 5, "Title", 48)

#endregion

#region Print all PDFs in a folder

dir c:\PowerShellDemo\*.pdf | foreach { Start-Process -FilePath $_.FullName Verb Print }

#endregion

#region changing units

#shows length in KB
dir $env:windir | select Mode, LastWriteTime, Length, Name | foreach { if ($_.Length -ne $null) { $_.Length = $_.Length | Add-Member ScriptMethod ToString { ('{0:0.0} KB' -f ($this/1KB)) } -Force -pass }; $_} | Sort-Object Length

#endregion

#region convert user to SID

function ConvertTo-SID($name, $domain=$env:userdomain) {
    $objUser = New-Object System.Security.Principal.NTAccount($domain, $name)
    $strSID = $objUser.Translate([System.Security.Principal.SecurityIdentifier])
    $strSID.Value
}

#endregion

#region HTML scraping

$webclient = New-Object System.Net.WebClient
$html = $webclient.DownloadString('http://drudgereport.com') | Out-String

$headerpattern = '(?i)<title>(.*?)</title>'

$header = ([regex]$headerpattern).Matches($html) |
  ForEach-Object { $_.Groups[1].Value }

$header

#endregion

#region text to speech

function Speak-Text {
    param(
        [Parameter(Mandatory=$true)]
        $Text
    )

    $object = New-Object -ComObject SAPI.SpVoice
    $object.Speak($text) | Out-Null
}

function Out-Speech {
  begin { $txt = "" }
  process { $txt += $_ }
  end { 
  	$object = New-Object -ComObject SAPI.SpVoice
    $object.Speak($txt) | Out-Null
	}
}
dir | select name | Out-String | Out-Speech

#endregion

#region ConvertFrom-String improvements

$template = @'
{Name*:Phoebe Cat}, {phone:425-123-6789}, {age:6}
{Name*:Lucky Shot}, {phone:(206) 987-4321}, {age:12}
'@

$testText = @'
Phoebe Cat, 425-123-6789, 6
Lucky Shot, (206) 987-4321, 12
Elephant Wise, 425-888-7766, 87
Wild Shrimp, (111)  222-3333, 1
'@

$testText | ConvertFrom-String -TemplateContent $template

#endregion

#region WMI

Get-wmiobject -list | where {$_.name -match "Network"}

Get-WMIObject Win32_BIOS

Get-WmiObject Win32_Environment

Get-WmiObject Win32_TimeZone

Get-WMIObject Win32_OperatingSystem

Get-WMIObject Win32_BootConfiguration

Get-WmiObject Win32_Process

Get-WMIObject Win32_OperatingSystem Property ServicePackMajorVersion

#endregion

#endregion Extra